#!/usr/bin/env python3

import numpy as np
import os


def test_filename(root, extension):
    """Returns a valid filename that does not already exist of the form
       <root> (N).<extension>, where N is the smallest possible number"""
    import glob

    existing = glob.glob("{}*.{}".format(root, extension))
    n = 2
    current = "{}.{}".format(root, extension)
    while current in existing:
        current = "{} ({}).{}".format(root, n, extension)
        n += 1
    return current


def txt_writer(input_file, input_columns, output_columns, common_inputs, multi_inputs):
    def output_formatter(val):
        if isinstance(val, (str, int, np.integer)):
            return str(val)
        if isinstance(val, bool):
            return str(val)
        if isinstance(val, (float, np.float)):
            return str(val)
        if isinstance(val, list):
            return "[" + ", ".join(output_formatter(value) for value in val) + "]"
        if isinstance(val, np.ma.core.MaskedConstant):
            return "---"
        if isinstance(val, np.ma.MaskedArray):
            arr = ["---" if m else a for a, m in zip(val, np.ma.getmaskarray(val))]
            return output_formatter(arr)
        if isinstance(val, np.ndarray):
            return output_formatter(list(val))
        raise TypeError("No formatting rule for type {}.".format(str(type(val))))

    output_string = list()
    output_string.append("Output of {}".format(input_file))
    output_string.append("# Common parameters:")
    for k, v in sorted(common_inputs.items()):
        v = output_formatter(v)
        output_string.append("# {:>15} = {}".format(k, v))
    no_experiments = len(multi_inputs)
    for i, mo in enumerate(multi_inputs):
        output_string.append("\n\n## Calculation {}/{}".format(i + 1, no_experiments))
        output_string.append("## Calculation specific variables:")
        selected_rows = [True] * len(output_columns['V'])
        for k, v in sorted(mo.items()):
            for j, select in enumerate(selected_rows[:]):
                selected_rows[j] = selected_rows[j] and v == input_columns[k][j]
            v = output_formatter(v)
            output_string.append("## {:>15} = {}".format(k, v))
        output_string.append("\n### Calculation results:")
        for k, v in sorted(output_columns.items()):
            v = [vv for sel, vv in zip(selected_rows, v) if sel]
            v = output_formatter(v)
            output_string.append("### {:>15} = {}".format(k, v))
    output_name = os.path.splitext(input_file)[0] + "-output"
    output_name = test_filename(output_name, "txt")
    print("Writing output (.txt) file [{}]".format(output_name))
    with open(output_name, mode="w", encoding="utf-8") as f:
        f.write("\n".join(output_string))


def dataframe_writer(input_file, input_columns, output_columns, common_inputs, multi_outputs):
    import pandas
    columns = dict()
    columns.update(input_columns)
    columns.update(output_columns)
    headings = sorted(input_columns.keys()) + sorted(output_columns.keys())
    df = pandas.DataFrame(columns, columns=headings)
    return df


def pd_writer(input_file, input_columns, output_columns, common_inputs, multi_outputs):
    df = dataframe_writer(input_file, input_columns, output_columns, common_inputs, multi_outputs)
    output_name = os.path.splitext(input_file)[0] + "-output"
    output_name = test_filename(output_name, "pd")
    print("Writing pandas (.pd) pickle file [{}]".format(output_name))
    df.to_pickle(output_name)


def xls_writer(input_file, input_columns, output_columns, common_inputs, multi_outputs):
    df = dataframe_writer(input_file, input_columns, output_columns, common_inputs, multi_outputs)
    output_name = os.path.splitext(input_file)[0] + "-output"
    output_name = test_filename(output_name, "xls")
    print("Writing excel (.xls) file [{}]".format(output_name))
    df.to_excel(output_name)

def np_writer(input_file, input_columns, output_columns, common_inputs, multi_outputs):
    import pickle
    columns = dict()
    columns.update(input_columns)
    columns.update(output_columns)
    keys = sorted(columns.keys())
    types = [type(columns[k][0]).__name__ for k in keys]
    types = ['unicode' if t == 'str' else t for t in types] # Workaround for unicode support of numpy
    matrix = [columns[k] for k in keys]
    dtypes = list(zip(keys, types))
    np_array = np.array(list(zip(*matrix)), dtype=dtypes)
    output_name = os.path.splitext(input_file)[0] + "-output"
    output_name = test_filename(output_name, "np")
    print("Writing numpy (.np) pickle file [{}]".format(output_name))
    with open(output_name, 'wb') as f:
        pickle.dump(np_array, f)


writers = {'txt': txt_writer,
           'pd': pd_writer,
           'np': np_writer,
           'xls': xls_writer}
